"""
The datatypes package contains abstract data types tailored for radio machine learning.
"""

__all__ = ["Annotation", "Recording"]

from .annotation import Annotation
from .recording import Recording
